/*
 * (C) Copyright 2001
 * Josh Huber <huber@mclx.com>, Mission Critical Linux, Inc.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * flash.c - 8 bit flash support for the 512k AMD on TSI108EMU/SVB 
 *           8 bit flash support for the 16MB 16bit extra flash on the SVB only.
 *           Most of this file was based on the existing U-Boot
 *           flash drivers.
 *
 * NOTE: In process of modification for boards mentioned above.
 *
 * Originally written or collected and sometimes rewritten by
 * Ingo Assmus <ingo.assmus@keymile.com>
 */

#include <common.h>
#include "tsi108.h"

#define HLP_Bx_CTRL0               (0x008)
#define HLP_Bx_CTRL0_FWE_MASK      (0x40000000)
#define HLP_Bx_CTRL0_WAIT_MASK     (0x3f000000)
#define HLP_Bx_CTRL0_THWR_MASK     (0x00000F00)
#define HLP_Bx_CTRL0_WIDTH_MASK    (0x00000003)

#define HLP_Bx_CTRL1               (0x00c)
#define HLP_Bx_CTRL1_ENABLE_MASK   (0x40000000)
#define HLP_Bx_CTRL1_WEOFF_MASK    (0x000f0000)

#define HLP_Bx_CTRL0_WAIT_VALUE    (0x1f000000)
#define HLP_Bx_CTRL0_THWR_VALUE    (0x00000600)  
#define HLP_Bx_CTRL1_WEOFF_VALUE   (0x00030000)  

#define HLP_WIDTH_8     (0x0)
#define HLP_WIDTH_16    (0x1)
#define HLP_WIDTH_32    (0x2)

#define FLASH_ROM       0xFFFD	/* unknown flash type                   */
#define FLASH_RAM       0xFFFE	/* unknown flash type                   */
#define FLASH_MAN_UNKNOWN 0xFFFF0000

#define FLASH_WORD_SIZE unsigned char

#define DONE32_2X16		(0x80008000) /* assumes LUT set to byte swapping */
#define DEBUG

flash_info_t flash_info[CFG_MAX_FLASH_BANKS];	/* info for FLASH chips */

/*-----------------------------------------------------------------------
 * Functions
 */
static ulong flash_get_size (int portwidth, vu_long * addr, 
                             flash_info_t * info);
static int write_word (flash_info_t * info, ulong dest, ulong data);
static void flash_get_offsets (ulong base, flash_info_t * info);

/*-----------------------------------------------------------------------
 */

unsigned long flash_init (void)
{
	short i;
	int   portwidth = 0;
	ulong size_b0 = 0;
	ulong hlp_ctrl0;

	/* Memory map currently supports one flash - will need to add more */
	/* code to support multiple/dynamic memory map/bank of flashs      */
	if (CFG_MAX_FLASH_BANKS > 1) {
		printf ("*** flash routines only support 1 bank of flashes!");
	}

	/* Init: no FLASHes known */
	for (i=0; i<CFG_MAX_FLASH_BANKS; i++)	{
		flash_info[i].flash_id = FLASH_UNKNOWN;
	}

	/* Re-configure HLP banks */
	/* 
			1.  Enable Writes
			2.  Reduce WAIT and WEOFF
			3.  Increase THWR
	*/
	i = 0;
	hlp_ctrl0 = *(ulong*)(CFG_TSI108_CSR_RST_BASE + TSI108_HLP_REG_OFFSET + 
		HLP_Bx_CTRL0 + (i*0x10));
	hlp_ctrl0 &= ~(HLP_Bx_CTRL0_FWE_MASK | 
		HLP_Bx_CTRL0_WAIT_MASK |
		HLP_Bx_CTRL0_THWR_MASK);
	hlp_ctrl0 |= (HLP_Bx_CTRL0_FWE_MASK | 
		HLP_Bx_CTRL0_WAIT_VALUE | 
		HLP_Bx_CTRL0_THWR_VALUE);

	*(ulong*)(CFG_TSI108_CSR_RST_BASE + TSI108_HLP_REG_OFFSET + 
		HLP_Bx_CTRL0 + (i*0x10)) = hlp_ctrl0;

	switch(hlp_ctrl0 & HLP_Bx_CTRL0_WIDTH_MASK) {
	default:
		portwidth = 1;
	break;
	case HLP_WIDTH_16:
		portwidth = 2;
		break;
	case HLP_WIDTH_32:
		/* Note */
		/* Timing and port width settings differs from default register settings */
		/* Need to initialize via EEPROM the following registers */
		/* mw c0000008 7ffc44c2 */
		/* mw c000000c 7c0f2000 */
		/* =>imd 50 0 40 */
		/* 0000: 00 02 ff ff ff ff ff ff 00 00 00 08 7f fc 44 c2 */
		/* 0010: 00 00 00 0c 7c 0f 20 00 ff ff ff ff ff ff ff ff */
		portwidth = 4;
		break;
	}

	asm("eieio");
	asm("sync");

	/* Initialize boot flash - assume it's the first flash */
	size_b0 = flash_get_size (portwidth, (vu_long *) CFG_FLASH_BASE, &flash_info[0]);

	printf ("[%ldkB@%lx] ", size_b0 / 1024, CFG_FLASH_BASE);

	if (flash_info[0].flash_id == FLASH_UNKNOWN) {
		printf ("## Unknown FLASH at %08lx: Size = 0x%08lx = %ld MB\n", CFG_FLASH_BASE, size_b0, size_b0 << 20);
	} else {
		/* Protect default boot image */
		switch (flash_info[0].flash_id & FLASH_TYPEMASK) {
		case FLASH_AM040:
			/* SVB board does not support AM29LV040B sector protection */
			break;
		case FLASH_28F128J3A:	
		/* There is 'real' chip protection for Intel*/
		/*
			*** Note ***
			CPU0 always boots with PB_OCN_BAR1[BOOT] == 1.
			Tsi108 will only use PB_A[16:35] and zeroes the
			remainder of the address.  This means a reset
			vector of 0xfff00100 is passed to flash as 0x0000_0100.

			Once the PB_OCN_BAR1[BOOT] == 0, the values in the
			corresponding LUTs take effect.  With the current
			memory map (16Meg LUT), subsequent accesses to address 
			0xfff00100 will be passed to flash as 0x00f00100.

			With the current dual CPU init sequence, CPU0 will clear the 
			PB_OCN_BAR1[BOOT] before enabling arbitration for CPU1.  This 
			means CPU1 reading of address 0xfff00100 will be passed to flash 
			as 0x00f00100 instead of 0x00000100 as was CPU0.

			For 512k flash (AMD29LV040B), this is okay because by default
			it will be treating 0x00f00100 as 0x00000100 anyway due since
			the maximum flash address accepted is 0x0007ffff.

			For 16Meg flash (Intel 28F128J3A), this will be a problem.
			Current workaround is to burn 2 identical boot images.
			CPU0 at sector 0:1     (corresponds to 0xff000000-0xff03ffff)
			CPU1 at sector 120:121 (corresponds to 0xfff00000-0xfff3ffff)

			Other possible workarounds:
			1.  Change the init sequence so that both CPUs runs with BOOT==1,
				so the initial flash access by the CPUs are consistent to the
				same flash location.

			**  Any soft reset after LUTs are set will have to enable
				boot bit again if LUTs do not mimic BOOT==1 translation
				to ensure flash access the same location as a cold boot.

			2.  Change LUTs to accomodate BOOT==1 translation of boot vector.
		*/

		/* Protect 1st image for CPU0 */
		(void)flash_protect(FLAG_PROTECT_SET, 
			(CFG_FLASH_BASE & 0xff000000),
			(CFG_FLASH_BASE & 0xff000000) + 0x40000 - 1,
			&flash_info[0]);
		/* Protect 2nd image for CPU1 */
		(void)flash_protect(FLAG_PROTECT_SET, 
			CFG_FLASH_BASE,
			CFG_FLASH_BASE + 0x40000 - 1,
			&flash_info[0]);
			break;
		default:
			printf ("Unknown Chip Type\n");
			break;
		}
	}

	return size_b0;
} /* end flash_init() */


/*-----------------------------------------------------------------------
 */
static void flash_get_offsets (ulong base, flash_info_t * info)
{
	int i;
	int sector_size;

	if (!info->sector_count)
		return;

	/* set up sector start address table */
	switch (info->flash_id & FLASH_TYPEMASK) {
	case FLASH_AM040:
	case FLASH_28F128J3A:	
	default:  /* For chips that have uniformly spaced sectors */
		sector_size = info->size / info->sector_count;
#ifdef DEBUG
		printf ("base 0x%08x, sector_size = %0d, size = %05x, sector_count = %d\n", base, sector_size, info->size, info->sector_count);
#endif // DEBUG
		for (i = 0; i < info->sector_count; i++) {
			info->start[i] = base + (i * sector_size);
		}
		break;
	}
} /* end flash_get_offsets() */

/*-----------------------------------------------------------------------
 */
void flash_print_info (flash_info_t * info)
{
	ulong i = 0;
	ulong hlp_ctrl0;
	int portwidth;

	/*
		Temporary fix, current memory map does not support 2 FLASH:devices  
		
		Allows manually switching between AMD/Intel on board flash  

		Procedure:
		mw c000000c bc0f2000   // Disable bank 0
		mw c0000010 0          // Map Bank 1 to 0
		mw c0000018 5ffc46c0   // Enable writes on Bank 1
		flinfo                 // Re-reads flash info
	*/
	hlp_ctrl0 = *(ulong*)(CFG_TSI108_CSR_RST_BASE + TSI108_HLP_REG_OFFSET + 
				HLP_Bx_CTRL0 + (i*0x10));

	switch(hlp_ctrl0 & HLP_Bx_CTRL0_WIDTH_MASK) {
	default:
		portwidth = 1;
	break;
	case HLP_WIDTH_16:
		portwidth = 2;
		break;
	case HLP_WIDTH_32:
		portwidth = 4;
		break;
	}
	flash_get_size (portwidth, (vu_long *) CFG_FLASH_BASE, &flash_info[0]);

	if (info->flash_id == FLASH_UNKNOWN) {
		printf ("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id & FLASH_VENDMASK) {
	case FLASH_MAN_STM:
		printf ("STM ");
		break;
	case FLASH_MAN_AMD:
		printf ("AMD ");
		break;
	case FLASH_MAN_FUJ:
		printf ("FUJITSU ");
		break;
	case FLASH_MAN_INTEL:
		printf ("INTEL ");
		break;
	default:
		printf ("Unknown Vendor ");
		break;
	}

	switch (info->flash_id & FLASH_TYPEMASK) {
	case FLASH_AM040:
		printf ("AM29LV040B (4 Mbit, bottom boot sect)\n");
		break;
	case FLASH_28F128J3A:	
		printf ("28F128J3A (128Mbit)\n");
		break;
	default:
		printf ("Unknown Chip Type\n");
		break;
	}

	if ((info->size >> 20) > 0) {
		printf ("  Size: %ld MB in %d Sectors\n",
			info->size >> 20, info->sector_count);
	} else {
		printf ("  Size: %ld kB in %d Sectors\n",
			info->size >> 10, info->sector_count);
	}

	printf ("  Sector Start Addresses:");
	for (i = 0; i < info->sector_count; i++) {
		if ((i % 5) == 0)
			printf ("\n   ");
		printf (" %08lX%s",
			info->start[i], info->protect[i] ? " (RO)" : "     ");
	}
	printf ("\n");
	return;
} /* end flash_print_info() */

/*-----------------------------------------------------------------------
 */


/*-----------------------------------------------------------------------
 */

/*
 * The following code cannot be run from FLASH!
 */

static inline void flash_cmd (int portwidth, volatile unsigned char *addr, int offset, 
                              unsigned char cmd)
{
	/* supports 1x8, 1x16, and 2x16 */
	/* 2x8 and 4x8 are not supported */
	if (4 == portwidth) {
		/* Assumptions: */
		/* 1.  END_MODE == Byte Swap in LUT (0xc00038f8 = 00000240) */
		/* 2.  2x16 chips */
		unsigned long cmd32 = (cmd << 24) | (cmd << 16) | (cmd << 8) | cmd;
		*(volatile unsigned long *) (addr + offset * 4) = cmd32;
	} else {
		/* 1x16 or 1x8 */
		*(volatile unsigned char *) (addr + offset) = cmd;
	}
	asm("eieio");
}

static ulong 
flash_get_size (int portwidth, vu_long * addr, 
				flash_info_t * info)
{	
	short i;
	unsigned short manuid=0, devid=0;
	volatile unsigned char *caddr = (unsigned char *)addr;
	char reset = 0xf0;
		
#ifdef DEBUG
	printf("DBG:  flash_get_size for address 0x%lx: \n", (unsigned long)caddr);
#endif

	/* Set size to zero to flag unknown flashes */
	info->size = 0;

	info->portwidth = portwidth;
#ifdef DEBUG
	printf("DBG: portwidth = %d\n", portwidth);
#endif

	/* Write auto select command: read Manufacturer ID */
	flash_cmd (portwidth, caddr, 0x0, 0xF0);   /* reset bank */

	flash_cmd (portwidth, caddr, 0x555, 0xAA);
	flash_cmd (portwidth, caddr, 0x2AA, 0x55);
	flash_cmd (portwidth, caddr, 0x555, 0x90);

	/* Read manufacturer ID and vendor ID */
	manuid = caddr[0];

	switch(manuid) {
	case (FLASH_WORD_SIZE)AMD_MANUFACT:
		info->flash_id = FLASH_MAN_AMD;
		reset          = 0xf0;
		devid          = caddr[1];
		asm ("eieio");

		switch(devid) {
		case (AMD_ID_LV040B):            
				info->flash_id     += FLASH_AM040;
				info->sector_count  = 8;
				info->size          = 0x80000;
				info->chipwidth     = 1;
				break;		          /* 512 kB */
		}
		break;
	case (FLASH_WORD_SIZE)INTEL_MANUFACT:
	case (FLASH_WORD_SIZE)INTEL_ALT_MANU:
		info->flash_id      = FLASH_MAN_INTEL;
		reset = 0xff;
		if(4 == info->portwidth)
			devid = caddr[4];
		else
			devid = caddr[2];  /* Note:  A0 is not used in either x8 or x16   */
							   /*        obtaining the id codes, lowest order */
							   /*        bit is A1.                           */
		asm ("eieio");

		switch(devid) {
		case (FLASH_WORD_SIZE)(INTEL_ID_28F128J3A):
			info->flash_id     += FLASH_28F128J3A;
			info->sector_count  = 128;
			info->size          = 0x1000000;
			info->chipwidth     = 2;
			caddr = (unsigned char *)0xff000000;  // vc temp
			break;		          /* 16 MB */
		}
		break;
	}

	#ifdef DEBUG
	printf("DBG:  Manufacturer: 0x%lx  devid 0x%1x\n", manuid, devid);
	#endif

	/* Get sector info */
	flash_get_offsets ((ulong)caddr, info);

	/* Check for protected sectors */
	for (i = 0; i < info->sector_count; i++) {
		/* read sector protection at sector address, (A7 .. A0) = 0x02 */
		/* D0 = 1 if protected */
		caddr = (volatile unsigned char *)(info->start[i]);

		switch((FLASH_VENDMASK)&info->flash_id) {
		case FLASH_MAN_INTEL:
			if(4 == portwidth)
				info->protect[i] = caddr[8] & 1;
			else
				info->protect[i] = caddr[4] & 1;
			break;
		case FLASH_MAN_AMD:
		default:
			info->protect[i] = caddr[2] & 1;
			break;
		}
	}

	if(0 == info->size)	{
		/* Treat unknown flash as ROM (ie. possible ROM emulator) */
		flash_cmd (portwidth, caddr, 0, 0xf0);
		printf ("Unknown Mfr [%08lx]:%08lx\n", manuid, devid);
		info->flash_id     = FLASH_UNKNOWN;
		info->sector_count = 1;
		info->size         = 0x80000;
		info->chipwidth    = 1;
	}

	/* Reset flash to read array mode */
	caddr = (volatile unsigned char *)(info->start[0]);
	flash_cmd (portwidth, caddr, 0x0, reset);
	
	return (info->size);
} /* end flash_get_size() */


int flash_erase (flash_info_t * info, int s_first, int s_last)
{
	volatile unsigned char *addr = (char *) (info->start[0]);
	int   flag, prot, sect, l_sect;
	ulong start, now, last;
	int   sector_size;
	char  status;
	int width = info->portwidth;

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return 1;
	}

	if (info->flash_id == FLASH_UNKNOWN) {  /* Flash is unknown */
		printf ("Can't erase unknown flash type %08lx - aborted\n", 
			info->flash_id);
		return 1;
	}

	prot = 0;
	for (sect = s_first; sect <= s_last; ++sect) {
		if (info->protect[sect]) {
			prot++;
		}
	}

	if (prot) {
		printf ("- Warning: %d protected sectors will not be erased!\n", prot);
	} else {
		printf ("\n");
	}

	l_sect = -1;

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts ();

	switch((FLASH_VENDMASK)&info->flash_id) {
	case FLASH_MAN_INTEL:
		/* Start erase on unprotected sectors */
		for (sect = s_first; sect <= s_last; sect++) {
			if (info->protect[sect] == 0) {	/* not protected */
				addr = (char *) (info->start[sect]);

				printf("Erasing sector %2d ... ", sect);

				start = get_timer (0);
				last  = start;

				flash_cmd (width, addr, 0, 0x50);	/* clear status register */
				flash_cmd (width, addr, 0, 0x20);	/* erase setup */
				flash_cmd (width, addr, 0, 0xd0);	/* erase confirm */

				if(4 == width) {
					ulong stat;
					volatile ulong *addr2 = (ulong *) (info->start[sect]);
					while (((stat = *addr2) & DONE32_2X16) != DONE32_2X16) {
						if ((now = get_timer (start)) > CFG_FLASH_ERASE_TOUT) {
							printf("Timeout\n");
							
							flash_cmd (width, addr, 0, 0xb0);
							flash_cmd (width, addr, 0, 0xff);
							return 1;
						}
										/* show that we're waiting */
						if ((now - last) > 1000) {	/* every second */
							putc ('.');
							last = now;
						}
					}
				} else {
					while (((status = *addr) & 0x80) != 0x80) {
						if ((now = get_timer (start)) > CFG_FLASH_ERASE_TOUT) {
							printf("Timeout\n");
							
							flash_cmd (width, addr, 0, 0xb0);
							flash_cmd (width, addr, 0, 0xff);
							return 1;
						}
										/* show that we're waiting */
						if ((now - last) > 1000) {	/* every second */
							putc ('.');
							last = now;
						}
					}
				}

				flash_cmd (width, addr, 0, 0x50);	/* clr stat reg cmd. */
				flash_cmd (width, addr, 0, 0xff); /* reset to read mode */

				printf(" done\n");
			}
		}
		break;
	case FLASH_MAN_AMD:
		default:
		flash_cmd (width, addr, 0x555, 0xAA);
		flash_cmd (width, addr, 0x2AA, 0x55);
		flash_cmd (width, addr, 0x555, 0x80);
		flash_cmd (width, addr, 0x555, 0xAA);
		flash_cmd (width, addr, 0x2AA, 0x55);

		/* Start erase on unprotected sectors */
		for (sect = s_first; sect <= s_last; sect++) {
			if (info->protect[sect] == 0) {	/* not protected */
				addr = (char *) (info->start[sect]);
				flash_cmd (width, addr, 0, 0x30);
				l_sect = sect;
			}
		}
		break;
	}

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts ();

	/* wait at least 80us - let's wait 1 ms */
	udelay (1000);

	/*
	 * We wait for the last triggered sector
	 */
	if (l_sect > 0)	{
		start = get_timer (0);
		last  = start;
		addr  = (volatile unsigned char *) (info->start[l_sect]);
		
		if(4 == width) {
			while ((addr[0] & DONE32_2X16) != DONE32_2X16) {
				if ((now = get_timer (start)) > CFG_FLASH_ERASE_TOUT) {
					printf ("Timeout\n");
					return 1;
				}
				/* show that we're waiting */
				if ((now - last) > 1000) {	/* every second */
					putc ('.');
					last = now;
				}
			}
		} else {
			while ((addr[0] & 0x80) != 0x80) {
				if ((now = get_timer (start)) > CFG_FLASH_ERASE_TOUT) {
					printf ("Timeout\n");
					return 1;
				}
				/* show that we're waiting */
				if ((now - last) > 1000) {	/* every second */
					putc ('.');
					last = now;
				}
			}
		}
	}

	/* reset to read array mode */
	addr = (volatile unsigned char *) info->start[0];
	flash_cmd (width, addr, 0x0, 0xf0);   /* reset bank */

	printf (" done \n");
	return 0;
} /* end flash_erase() */

/*-----------------------------------------------------------------------
* Copy memory to flash, returns:
* 0 - OK
* 1 - write timeout
* 2 - Flash not erased
*/

/* vc broken for 2x16: TODO */
int write_buff (flash_info_t * info, uchar * src, ulong addr, ulong cnt)
{
	ulong cp, wp, data;
	int i, l, rc;

	wp = (addr & ~3);	/* get lower word aligned address */

	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
		data = 0;
		for (i = 0, cp = wp; i < l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *) cp);
		}
		for (; i < 4 && cnt > 0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt == 0 && i < 4; ++i, ++cp) {
			data = (data << 8) | (*(uchar *) cp);
		}

		if ((rc = write_word (info, wp, data)) != 0) {
			return (rc);
		}
		wp += 4;
	}

	/*
	 * handle word aligned part
	 */
	while (cnt >= 4) {
		data = 0;
		for (i = 0; i < 4; ++i) {
			data = (data << 8) | *src++;
		}
		if ((rc = write_word (info, wp, data)) != 0) {
			return (rc);
		}
		wp += 4;
		cnt -= 4;
	}

	if (cnt == 0) {
		return (0);
	}

	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i = 0, cp = wp; i < 4 && cnt > 0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}
	for (; i < 4; ++i, ++cp) {
		data = (data << 8) | (*(uchar *) cp);
	}

	return (write_word (info, wp, data));
} /* end write_buff() */

/*-----------------------------------------------------------------------
* Write a word to Flash, returns:
* 0 - OK
* 1 - write timeout
* 2 - Flash not erased
*/

static int write_word (flash_info_t * info, ulong dest, ulong data)
{
	volatile unsigned char *addr = (char *) (info->start[0]);
	ulong start, now, last;
	int flag, i;
	char status;
	int width = info->portwidth;

	/* Check if Flash is (sufficiently) erased */
	if ((*((vu_long *) dest) & data) != data) {
		ulong tmp;
		tmp = *(vu_long *)dest;
		*addr = 0xFFFFFFFF;	/* resest to read mode          */
		return (2);
	}

	/* Disable interrupts to prevent possible timeout*/
	flag = disable_interrupts ();

	switch((FLASH_VENDMASK)&info->flash_id) {
	case FLASH_MAN_INTEL:

		if(4 == width) {
			ulong stat;
			volatile ulong *addr2 = (ulong *) (info->start[0]);

			/* Word/Byte Program */
			/* Assumes END_MODE == Byte Swap in LUT (0xc00038f8 = 00000240) */
				flash_cmd (width, addr, 0x00, 0x40);	
				*(ulong*)dest = data;

				start = get_timer (0);
				last  = start;

				while (((stat = *addr2) & DONE32_2X16) != DONE32_2X16) {
					if ((now = get_timer (start)) > (0x120000)) {
						printf("Timeout\n");
						*addr2 = 0xFFFFFFFF;	/* resest to read mode */
						return 1;
					}
				}
		} else {
			/* Program byte */
			for (i = 0; i < 4; i++) {
				char *data_ch = (char *) &data;

				/* Word/Byte Program */
				flash_cmd (width, addr, 0x00, 0x40);	
				flash_cmd (width, ((char *)(dest+i)), 0x00, data_ch[i]);

				start = get_timer (0);
				last  = start;

				while (((status = *addr) & 0x80) != 0x80) {
					if ((now = get_timer (start)) > 0x120000) {
						printf("Timeout\n");
						*addr = 0xFF;	/* resest to read mode          */
						return 1;
					}
				}
			}
		}

		flash_cmd (width, addr, 0x0, 0x50);   /* clear status register cmd.   */
		flash_cmd (width, addr, 0x0, 0xff);   /* reset to read mode          */
		break;

	case FLASH_MAN_AMD:
	default:
		/* Unlock bypass command */
		flash_cmd (width, addr, 0x555, 0xAA);
		flash_cmd (width, addr, 0x2AA, 0x55);
		flash_cmd (width, addr, 0x555, 0x20);	

		/* Program byte */
		for (i = 0; i < 4; i++) {
			char *data_ch = (char *) &data;

			/* Unlock Bypass Program */
			flash_cmd (width, addr, 0x000, 0xA0);	
			flash_cmd (width, ((char *)(dest+i)), 0x00, data_ch[i]);
			udelay (15);	/* Spec says ~9us to program, udelay not that accurate, -~2-3us */
		}

		/* Unlock Bypass Reset */
		flash_cmd (width, addr, 0x0, 0x90);	
		flash_cmd (width, addr, 0x0, 0x00);	
		break;
	}

	/* Re-enable interrupts */
	if (flag)
		enable_interrupts ();

	/* data polling for D7 */
	start = get_timer (0);
	while ((*((vu_long *) dest) & 0x00800080) != (data & 0x00800080)) {
		if (get_timer (start) > CFG_FLASH_WRITE_TOUT) {
			return (1);
		}
	}
	return (0);
} /* end write_word() */

static int _flash_protect(flash_info_t *info, long sector)
{
	int i;
	int flag;
	ulong status;
	int rcode = 0;
	int width = info->portwidth;
	volatile unsigned char *addr = (unsigned char *)sector;

	switch(info->flash_id & FLASH_TYPEMASK) {

	case FLASH_28F128J3A:
		/* Disable interrupts which might cause Flash to timeout */
		flag = disable_interrupts ();

		/* Issue command */
		flash_cmd (width, addr, 0, 0x50);	/* clear status register */
		flash_cmd (width, addr, 0, 0x60);	/* set lock bit setup */
		flash_cmd (width, addr, 0, 0x01);	/* set lock bit confirm */

		/* Wait for command completion */
		for (i = 0; i < 10; i++) {
			udelay(100);
			if(4 == info->portwidth) {
				if ((*addr & DONE32_2X16) == DONE32_2X16)
					break;
			} else {
				if ((*addr & 0x80) == 0x80)
					break;
			}
		}

		/* Not successful? */
		if(4 == info->portwidth) {
			status = *(ulong *)(sector);
			if ((status & DONE32_2X16) != DONE32_2X16) {
				printf("Protect %x sector failed: %08x\n", (uint)sector, (uint)status);
				rcode = 1;
			}
		} else {
			status = *addr;
			if ((status & 0x80) != 0x80) {
				printf("Protect %x sector failed: %x\n", (uint)sector, (uint)status);
				rcode = 1;
			}
		}

		/* restore read mode */
		flash_cmd (width, addr, 0, 0xff);	/* Restore read mode*/

		/* re-enable interrupts if necessary */
		if (flag)
			enable_interrupts();

		break;

	case FLASH_AM040:                    /* No soft sector protection */
		printf("SVB does not support sector prot/unprot for AM29LV040B\n");
		break;
	}

	/* Turn protection on for this sector */
	for (i = 0; i < info->sector_count; i++) {
		if (info->start[i] == sector) {
			info->protect[i] = 1;
			break;
		}
	}

	return rcode;
} /* end _flash_protect() */

static int _flash_unprotect(flash_info_t *info, long sector)
{
	int i;
	int flag;
	ulong status;
	int rcode = 0;
	int width = info->portwidth;
	volatile unsigned char *addr = (unsigned char *)sector;

	switch(info->flash_id & FLASH_TYPEMASK) {
	case FLASH_28F128J3A:
		/* Disable interrupts which might cause Flash to timeout */
		flag = disable_interrupts ();

		flash_cmd (width, addr, 0, 0x50);	/* clear status register */
		flash_cmd (width, addr, 0, 0x60);	/* clear lock bit setup */
		flash_cmd (width, addr, 0, 0xd0);	/* clear lock bit confirm */

		/* Wait for command completion */
		for (i = 0; i < 100 ; i++) { /* 700ms timeout, wait > 700ms */
			udelay(100000);			 /* Delay 100ms */
			if(4 == info->portwidth) {
				if ((*addr & DONE32_2X16) == DONE32_2X16)
					break;
			} else {
				if ((*addr & 0x80) == 0x80)
					break;
			}
		}

		/* Not successful? */
		if(4 == info->portwidth) {
			status = *(ulong *)(sector);
			if ((status & DONE32_2X16) != DONE32_2X16) {
				printf("Un-protect %x sector failed: %08x\n", (uint)sector, (uint)status);
				rcode = 1;
			}
		} else {
			status = *addr;
			if ((status & 0x80) != 0x80) {
				printf("Un-protect %x sector failed: %x\n", (uint)sector, (uint)status);
				rcode = 1;
			}
		}

		/* restore read mode */
		flash_cmd (width, addr, 0, 0xff);	/* Restore read mode*/
		
		/* re-enable interrupts if necessary */
		if (flag)
			enable_interrupts();

		break;
	case FLASH_AM040:                    /* No soft sector protection */
		printf("SVB does not support sector prot/unprot for AM29LV040B\n");
		break;
	}

	/*
	 * Fix Intel's little red wagon.  Reprotect
	 * sectors that were protected before we undid
	 * protection on a specific sector.
	 */
	for (i = 0; i < info->sector_count; i++) {
		if (info->start[i] != sector) {
			if (info->protect[i]) {
				if (_flash_protect(info, info->start[i]))
					rcode = 1;
			}
		} else {
			/* Turn protection off for this sector */
			info->protect[i] = 0;
		}
	}

	return rcode;
} /* end _flash_unprotect() */


int flash_real_protect(flash_info_t *info, long sector, int prot)
{
	int rcode;

	if (prot)
		rcode = _flash_protect(info, info->start[sector]);
	else
		rcode = _flash_unprotect(info, info->start[sector]);

	return rcode;
} /* end flash_real_protect() */
