// Driver for the  TSI108 Gigabit Ethernet.
// Currently, only Marvell 88E1020 PHYs are supported.
// 
// Copyright (C) 2005 Tundra Semiconductor
// Kong Lai, <kong.lai@tundra.com>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation.

#include <common.h>
#include <malloc.h>
#include <net.h>
#include <asm/io.h>
//#include <tsi108.h>
#include "tsi108_eth.h"
#include <asm-ppc/errno.h>

#define ASSERT(expr, func) if( !( expr ) ) { \
		printf( "Assertion failed! %s:line %d %s\n", \
		(int)__FUNCTION__,__LINE__,(int)(#expr),0,0,0); \
		func }
#undef ETH_DBG

// Do not set Rx Desc Ring length to less than 32
#define TSI108_RXRING_LEN     32

// NOTE: The driver currently does not support receiving packets
// larger than the buffer size, so don't decrease this (unless you
// want to add such support).
#define TSI108_RXBUF_SIZE     1536

// Do not set Tx Desc Ring length to less than 32
#define TSI108_TXRING_LEN     32

typedef struct {
  volatile u32 *regs;    // Base of normal regs
  volatile u32 *phyregs; // Base of register bank used for PHY access
  int phy;               // Index of PHY for this interface

  // Next entry in rxring to read
  int rxtail;
	
  // Next entry in rxring to give a new buffer
  int rxhead;
	
  // Number of free, allocated RX buffers
  int rxfree;
	
  // Non-zero if there are still descriptors to be processed
  // from a previous descriptor interrupt condition that
  // has been cleared.
	
  int rxpending;

  // Next TX descriptor to check status on
  int txtail;
	
  // Next TX descriptor to use
  int txhead;
	
  // Number of free TX descriptors.  This could be calculated from
  // rxhead and rxtail if one descriptor were left unused to disambiguate
  // full and empty conditions, but it's simpler to just keep track
  // explicitly.
  int txfree;
	
  // Driver has a valid MAC address.  If zero, driver will not
  // contaminate the network with its packets.
  int mac_ok;
	
  // The PHY is currently powered on.
  int phy_ok;
	
  // PHY status (duplex is 1 for half, 2 for full,
  // so that the default 0 indicates that neither has
  // yet been configured).

  int link_up, speed, duplex;

  //memory for descriptor rings are allocated in tsi108_ifup during init
  tx_desc *txring;
  rx_desc *rxring;

  unsigned char *tx_buff[TSI108_RXRING_LEN];
  unsigned char *rx_buff[TSI108_RXRING_LEN];
	
  unsigned int txdma, rxdma;

  u32 txphys, rxphys;
	
  // These stats are kept separate in hardware, thus require individual
  // fields for handling carry.  They are currently not available.
  unsigned long rx_fcs;        // Add to rx_frame_errors
  unsigned long rx_short_fcs;  // Add to rx_frame_errors
  unsigned long rx_long_fcs;   // Add to rx_frame_errors
  unsigned long rx_underruns;  // Add to rx_length_errors
  unsigned long rx_overruns;   // Add to rx_length_errors

  unsigned long tx_coll_abort; // Add to tx_aborted_errors and collisions
  unsigned long tx_pause_drop; // Add to tx_aborted_errors
	
  u32 mc_hash[16];
} drv_data;

static struct eth_device *tsi108_devs[2] = {NULL, NULL};

static unsigned char mac_addr[2][6] = {{0x0,0x06,0xd2,0x0,0x0,0x1},
				       {0x0,0x06,0xd2,0x0,0x0,0x2}};
#if 0
static void dump_eth_one(struct eth_device *dev)
{
	drv_data *data = dev->priv;
	
	printf("Dumping %s...\n", dev->name);
	printf("intstat %x intmask %x mac_ok %d phy_ok %d link %d speed %d duplex %d\n", 
	       data->regs[TSI108_EC_INTSTAT], 
	       data->regs[TSI108_EC_INTMASK],
	       data->mac_ok, data->phy_ok,
	       data->link_up, data->speed, data->duplex);

	printf("TX: head %d, tail %d, free %d, stat %x, estat %x, err %x\n",
	       data->txhead, data->txtail, data->txfree,
	       data->regs[TSI108_EC_TXSTAT],
	       data->regs[TSI108_EC_TXESTAT],
	       data->regs[TSI108_EC_TXERR]);
	
	printf("RX: head %d, tail %d, free %d, stat %x, estat %x, err %x, pending %d\n\n",
	       data->rxhead, data->rxtail, data->rxfree,
	       data->regs[TSI108_EC_RXSTAT],
	       data->regs[TSI108_EC_RXESTAT],
	       data->regs[TSI108_EC_RXERR],
	       data->rxpending);
}

void tsi108_dump_eth(void)
{
	dump_eth_one(tsi108_devs[0]);
	dump_eth_one(tsi108_devs[1]);
}
#endif

static inline u16 tsi108_read_mii(drv_data *data, int reg, int *status)
{
	int i;
	u16 ret;

	data->phyregs[TSI108_MAC_MII_ADDR] = (data->phy << TSI108_MAC_MII_ADDR_PHY)
	  | (reg << TSI108_MAC_MII_ADDR_REG);

	data->phyregs[TSI108_MAC_MII_CMD] = 0;

	sync();
	
	data->phyregs[TSI108_MAC_MII_CMD] = TSI108_MAC_MII_CMD_READ;

	sync();
	udelay(10);
	
	for (i = 0; i < 100; i++) {
		if (!(data->phyregs[TSI108_MAC_MII_IND] & 
		      (TSI108_MAC_MII_IND_NOTVALID | TSI108_MAC_MII_IND_BUSY)))

			break;

		udelay(10);
	}
	
	if (i == 100) {
		if (status)
			*status = -EBUSY;
		
		ret = 0xffff;
	} else {
		if (status)
	 		*status = 0;
	 	
	 	ret = data->phyregs[TSI108_MAC_MII_DATAIN];
	}

	sync();
	
	return ret;
}

static inline void tsi108_write_mii(drv_data *data, int reg, u16 val)
{
	data->phyregs[TSI108_MAC_MII_ADDR] = (data->phy << TSI108_MAC_MII_ADDR_PHY)
	                               | (reg << TSI108_MAC_MII_ADDR_REG);
	sync();
	data->phyregs[TSI108_MAC_MII_DATAOUT] = val;
	sync();
	while (data->phyregs[TSI108_MAC_MII_IND] & TSI108_MAC_MII_IND_BUSY);
}

static inline void tsi108_write_tbi(drv_data *data, int reg, u16 val)
{
	data->regs[TSI108_MAC_MII_ADDR] = (0x1e << TSI108_MAC_MII_ADDR_PHY)
	  | (reg << TSI108_MAC_MII_ADDR_REG);
	sync();

#ifdef ETH_DBG
	printf("MII Mgmt Configuration Reg.: %08x\nMII Mgmt Command Reg.: %08x\nMII Mgmt Addr Reg.: %08x\nMII Mgmt Ctrl Reg.: %08x\nMII Mgmt Status Reg.: %08x\nMII Mgmt Indicators Reg.: %08x\n", data->regs[TSI108_MAC_MII_MGMT_CFG],
	       data->regs[TSI108_MAC_MII_CMD],
	       data->regs[TSI108_MAC_MII_ADDR],
	       data->regs[TSI108_MAC_MII_DATAOUT],
	       data->regs[TSI108_MAC_MII_DATAIN],
	       data->regs[TSI108_MAC_MII_IND]);
#endif
	data->regs[TSI108_MAC_MII_DATAOUT] = (u32)val;
	sync();
	while (data->regs[TSI108_MAC_MII_IND] & TSI108_MAC_MII_IND_BUSY);
}

static void tsi108_check_phy(struct eth_device *dev)
{
	drv_data *data = dev->priv;
	u16 sumstat;
	volatile u16 tmp = 0;

	// Do a dummy read, as for some reason the first read
	// after a link becomes up returns link down, even if
	// it's been a while since the link came up.

	if (!data->phy_ok) {
#ifdef ETH_DBG
	  printf("!data->phy_ok, returning early!\n");
#endif
	  return;
	}



	if (!((tmp = tsi108_read_mii(data, PHY_STAT, NULL)) & PHY_STAT_LINKUP)) {
	  if (data->link_up == 1) {
	    data->link_up = 0; 
	  }
		
#ifdef ETH_DBG
	  printf("Returning early. PHY_STAT: %04x\n", tmp);
#endif
	  return;
	}

	sumstat = tsi108_read_mii(data, PHY_SUM_STAT, NULL);

#ifdef ETH_DBG
	printf("sumstat: %04x\n", sumstat);
#endif

#ifdef CONFIG_TSI108SVB		/* SV board */
	switch (sumstat & PHY_SUM_STAT_SPEED_MASK) {
		case PHY_SUM_STAT_SPEED_10:
			if (data->speed != 10) {
				data->regs[TSI108_MAC_CFG2] &= ~TSI108_MAC_CFG2_IFACE_MASK;
				data->regs[TSI108_MAC_CFG2] |= TSI108_MAC_CFG2_NOGIG;
				data->regs[TSI108_EC_PORTCTRL] |= TSI108_EC_PORTCTRL_NOGIG;
				data->speed = 10;
			}
			break;
		
		case PHY_SUM_STAT_SPEED_100:
			if (data->speed != 100) {
				data->regs[TSI108_MAC_CFG2] &= ~TSI108_MAC_CFG2_IFACE_MASK;
				data->regs[TSI108_MAC_CFG2] |= TSI108_MAC_CFG2_NOGIG;
				data->regs[TSI108_EC_PORTCTRL] |= TSI108_EC_PORTCTRL_NOGIG;
				data->speed = 100;
			}
			break;
		
		case PHY_SUM_STAT_SPEED_1000:
			if (data->speed != 1000) {
				data->regs[TSI108_MAC_CFG2] &= ~TSI108_MAC_CFG2_IFACE_MASK;
				data->regs[TSI108_MAC_CFG2] |= TSI108_MAC_CFG2_GIG;
				data->regs[TSI108_EC_PORTCTRL] &= ~TSI108_EC_PORTCTRL_NOGIG;
				data->speed = 1000;
			}
			break;

		default:
#ifdef ETH_DBG
		  printf("PHY reported invalid speed.\n");
#endif
		  return;
	}

	if (sumstat & PHY_SUM_STAT_FULLDUPLEX) {
		if (data->duplex != 2) {
			data->regs[TSI108_MAC_CFG2] |= TSI108_MAC_CFG2_FULLDUPLEX;
			data->regs[TSI108_EC_PORTCTRL] &= ~TSI108_EC_PORTCTRL_HALFDUPLEX;
			data->duplex = 2;
		}
	} else {
		if (data->duplex != 1) {
			data->regs[TSI108_MAC_CFG2] &= ~TSI108_MAC_CFG2_FULLDUPLEX;
			data->regs[TSI108_EC_PORTCTRL] |= TSI108_EC_PORTCTRL_HALFDUPLEX;
			data->duplex = 1;
		}
	}

#elif defined(CONFIG_TSI108EVB)	/* EV board (very close to SVB)*/
	{
	u32 mac_cfg2_reg, portctrl_reg;
	u32 fdx_flag = 0, reg_update = 0;

	mac_cfg2_reg = data->regs[TSI108_MAC_CFG2];
	portctrl_reg = data->regs[TSI108_EC_PORTCTRL];

	switch (sumstat & PHY_SUM_STAT_SPEED_MASK) {
		case PHY_SUM_STAT_1000T_FD:
			fdx_flag++;
		case PHY_SUM_STAT_1000T_HD:
			if (data->speed != 1000) {
				mac_cfg2_reg &= ~TSI108_MAC_CFG2_IFACE_MASK;
				mac_cfg2_reg |= TSI108_MAC_CFG2_GIG;
				portctrl_reg &= ~TSI108_EC_PORTCTRL_NOGIG;
				data->speed = 1000;
				reg_update++;
			}
			break;


		case PHY_SUM_STAT_100TX_FD:
			fdx_flag++;
		case PHY_SUM_STAT_100TX_HD:
			if (data->speed != 100) {
				mac_cfg2_reg &= ~TSI108_MAC_CFG2_IFACE_MASK;
				mac_cfg2_reg |= TSI108_MAC_CFG2_NOGIG;
				portctrl_reg |= TSI108_EC_PORTCTRL_NOGIG;
				data->speed = 100;
				reg_update++;
			}
			break;
		
		case PHY_SUM_STAT_10T_FD:
			fdx_flag++;
		case PHY_SUM_STAT_10T_HD:
			if (data->speed != 10) {
				mac_cfg2_reg &= ~TSI108_MAC_CFG2_IFACE_MASK;
				mac_cfg2_reg |= TSI108_MAC_CFG2_NOGIG;
				portctrl_reg |= TSI108_EC_PORTCTRL_NOGIG;
				data->speed = 10;
				reg_update++;
			}
			break;

		default:
#ifdef ETH_DBG
		  printf("PHY reported invalid speed.\n");
#endif
			return;
	}

	if (fdx_flag) {
		if (data->duplex != 2) {
			mac_cfg2_reg |= TSI108_MAC_CFG2_FULLDUPLEX;
			portctrl_reg &= ~TSI108_EC_PORTCTRL_HALFDUPLEX;
			data->duplex = 2;
			reg_update++;
		}
	} else {
		if (data->duplex != 1) {
			mac_cfg2_reg &= ~TSI108_MAC_CFG2_FULLDUPLEX;
			portctrl_reg |= TSI108_EC_PORTCTRL_HALFDUPLEX;
			data->duplex = 1;
			reg_update++;
		}
	}

	if (reg_update) {
		data->regs[TSI108_MAC_CFG2] = mac_cfg2_reg;
		sync();
		data->regs[TSI108_EC_PORTCTRL] = portctrl_reg;
		sync();
	}

	}
#endif /* CONFIG_TSI108SVB */

	if (data->link_up == 0) {
		// The manual says it can take 3-4 usecs for the speed change
		// to take effect.
		udelay(5);
		data->link_up = 1;
	}
}

static void tsi108_restart_rx(drv_data *data, struct eth_device *dev)
{
  data->regs[TSI108_EC_RXQ_PTRHIGH] = TSI108_EC_RXQ_PTRHIGH_VALID;

  sync();
  data->regs[TSI108_EC_RXCTRL] = TSI108_EC_RXCTRL_GO | TSI108_EC_RXCTRL_QUEUE0;
}

static void tsi108_restart_tx(drv_data *data)
{
  data->regs[TSI108_EC_TXQ_PTRHIGH] = TSI108_EC_TXQ_PTRHIGH_VALID;

  sync();
  data->regs[TSI108_EC_TXCTRL] = TSI108_EC_TXCTRL_IDLEINT |
    TSI108_EC_TXCTRL_GO |
    TSI108_EC_TXCTRL_QUEUE0;
}

static void tsi108_check_for_completed_tx(struct eth_device *dev)
{
	drv_data *data = dev->priv;

	//data->txfree == 0 means the system has used up the entire
	// tx desc ring and is waiting for GIGE Tx DMA to send the
	// packets out.
	while (!data->txfree || data->txhead != data->txtail) {
		int tx = data->txtail;
		int last = 0;
		unsigned char * buff_ptr;
	
		if (data->txring[tx].misc & TSI108_TX_OWN) {
#ifdef ETH_DBG
		  printf("data->txring[tx].misc is %08x\n",
			 data->txring[tx].misc);
#endif
		  break;
		}
		
		buff_ptr = data->tx_buff[tx];
		
		if (!(data->txring[tx].misc & TSI108_TX_OK))
			printf("%s: bad tx packet, misc %x\n", dev->name, data->txring[tx].misc);

		// No more descriptor interrupts, since the queue is no longer
		// full.  Idle interrupts will still occur.

		data->regs[TSI108_EC_INTMASK] |= TSI108_INT_TXQUEUE0;

		data->txtail = (data->txtail + 1) % TSI108_TXRING_LEN;
		data->txfree++;
		
		if (data->txring[tx].misc & TSI108_TX_EOF) {
			last = 1;
		}

		if (last) {
#ifdef ETH_DBG
		  printf("Tx completed, free buff_ptr!\n");
#endif
		  free(buff_ptr);
		}
	}
#ifdef ETH_DBG
	printf("data->txfree is %08x\n", data->txfree);
#endif
}

static int tsi108_send_packet(struct eth_device *dev,
			      volatile void *packet,
			      int length)
{
        drv_data *data = dev->priv;
	unsigned char *tx_buffer = (unsigned char *)malloc(TSI108_RXBUF_SIZE);
	//We want to copy the entire packet here. That way, we don't have to
	// worry about when higher-level application frees its buffer.
	memcpy((char *)tx_buffer, (char *)packet, (int)length);
	
	//PADDING AND CRC ARE ADDED BY MAC
	
	if (!data->link_up) {
		printf("%s: link down!\n", dev->name);
		return 1;
	}

	if (!data->mac_ok) {
		printf("%s: Bad MAC!\n", dev->name);
		return 1;
	}
	
	if (!(data->regs[TSI108_EC_TXSTAT] & TSI108_EC_TXSTAT_QUEUE0))
	  tsi108_restart_tx(data);
	


	int misc = 0;
	int tx = data->txhead;
	
	data->tx_buff[tx] = tx_buffer;
	data->txring[tx].buf0 = (u32)tx_buffer;
	data->txring[tx].len = (u16)length;
	misc |= TSI108_TX_SOF | TSI108_TX_EOF;
	sync();
	data->txring[tx].misc = misc | TSI108_TX_OWN;
	data->txhead = (data->txhead + 1) % TSI108_TXRING_LEN;
	data->txfree--;

#ifdef ETH_DBG
	printf("AFTER\n");
	printf("txQ ptr low is %08x\n",
	       data->regs[TSI108_EC_TXQ_PTRLOW]);
	printf("txQ ptr high is %08x\n",
	       data->regs[TSI108_EC_TXQ_PTRHIGH]);
	printf("tx control: %08x\n",
	       data->regs[TSI108_EC_TXCTRL]);
	printf("tx status: %08x\n",
	       data->regs[TSI108_EC_TXSTAT]);
	printf("interrupt status: %08x\n",
	       data->regs[TSI108_EC_INTSTAT]);
#endif

	tsi108_check_for_completed_tx(dev);

	// This must be done after the check for completed tx descriptors,
	// so that the tail pointer is correct.

	if (!(data->regs[TSI108_EC_TXSTAT] & TSI108_EC_TXSTAT_QUEUE0))
		tsi108_restart_tx(data);
	
	return 0;
}

static int tsi108_check_for_completed_rx(struct eth_device *dev)
{
        drv_data *data = dev->priv;
	int done = 0;

	//data->rxfree starts at TSI108_RXRING_LEN
	while (data->rxfree) {
	  int rx = data->rxtail;
	  unsigned char * skb;

	  if (data->rxring[rx].misc & TSI108_RX_OWN)
	    break;
		
	  skb = data->rx_buff[rx];
	  data->rxtail = (data->rxtail + 1) % TSI108_RXRING_LEN;
	  data->rxfree--;
	  done++;

	  if (data->rxring[rx].misc & TSI108_RX_BAD) {

	    if (data->rxring[rx].misc & TSI108_RX_CRC)
	      //FIXME: do nothing for now

	    if (data->rxring[rx].misc & TSI108_RX_OVER)
	      //FIXME: do nothing for now

	      //free rx buffer
	      free(skb);

	    data->rxring[rx].blen = 1536;
	    data->rxfree++;

	    //There is no need to keep track of data->rxhead since
	    // rx buffers will be reused.
	  
	    data->rxring[rx].misc = TSI108_RX_OWN | TSI108_RX_INT;

	    sync();
	    continue;
	  }

	  NetReceive(skb, data->rxring[rx].len);

	  //make rx desc and buffer available to Rx DMA again
	  data->rxring[rx].blen = 1536;
	  data->rxfree++;
	  
	  data->rxring[rx].misc = TSI108_RX_OWN | TSI108_RX_INT;

	  sync();
	  
	}
	
	return done;
}

static int tsi108_poll(struct eth_device *dev)
{
	drv_data *data = dev->priv;
	u32 estat = data->regs[TSI108_EC_RXESTAT];
	u32 intstat = data->regs[TSI108_EC_INTSTAT];
	int num_received = 0;

	if (!(data->regs[TSI108_EC_RXSTAT] & TSI108_EC_RXSTAT_QUEUE0))
	  tsi108_restart_rx(data, dev);

	intstat &= TSI108_INT_RXQUEUE0 | TSI108_INT_RXTHRESH |
	           TSI108_INT_RXOVERRUN | TSI108_INT_RXERROR |
	           TSI108_INT_RXWAIT;

	data->regs[TSI108_EC_RXESTAT] = estat;
	data->regs[TSI108_EC_INTSTAT] = intstat;

	if (estat & TSI108_EC_RXESTAT_Q0_DESCINT)
	  num_received = tsi108_check_for_completed_rx(dev);
	
	//For U-BOOT, just call NetReceive and give the ownership of the
	// descriptot back to GIGE Rx DMA. 

	if (intstat & TSI108_INT_RXERROR) {
	  u32 err = data->regs[TSI108_EC_RXERR];
	  data->regs[TSI108_EC_RXERR] = err;

	  if (err)
	    printf("RX error status: %x\n", err);
	}

	if (intstat & TSI108_INT_RXOVERRUN) {

	}
	
	return 1;
}

static void tsi108_stop_ethernet(struct eth_device *dev)
{
#ifdef ETH_DBG
  printf("Halt does not disable TSI108 ethernet port!\n");
#endif
#if 0
  unsigned short i;
  drv_data *data = dev->priv;

  if (data != NULL) {
   // Disable all TX and RX queues...

   //data->regs[TSI108_EC_TXCTRL] = 0;
   //data->regs[TSI108_EC_RXCTRL] = 0;

   // ...and wait for them to become idle.
	
   //sync();
   //while (data->regs[TSI108_EC_TXSTAT] & TSI108_EC_TXSTAT_ACTIVE);
   //while (data->regs[TSI108_EC_RXSTAT] & TSI108_EC_RXSTAT_ACTIVE);
   
   //for (i = 0; i < TSI108_RXRING_LEN; i++) {
   //  free(data->rx_buff[i]);
   //  data->rx_buff[i] = NULL;
   //
   //  printf("data->rxring+i: %08x\n", (unsigned int)(data->rxring + i));
   //  free(data->rxring + i);
   //  free(data->txring + i);
   //}

  }
#endif
}


static void tsi108_reset_ether(drv_data *data)
{
	data->regs[TSI108_MAC_CFG1] = TSI108_MAC_CFG1_SOFTRST;
	udelay(100);
	data->regs[TSI108_MAC_CFG1] = 0;
	
	data->regs[TSI108_EC_PORTCTRL] = TSI108_EC_PORTCTRL_STATRST;
	udelay(100);
	data->regs[TSI108_EC_PORTCTRL] &= ~TSI108_EC_PORTCTRL_STATRST;
	
	data->regs[TSI108_EC_TXCFG] = TSI108_EC_TXCFG_RST;
	udelay(100);
	data->regs[TSI108_EC_TXCFG] &= ~TSI108_EC_TXCFG_RST;

	data->regs[TSI108_EC_RXCFG] = TSI108_EC_RXCFG_RST;
	udelay(100);
	data->regs[TSI108_EC_RXCFG] &= ~TSI108_EC_RXCFG_RST;

	data->regs[TSI108_MAC_MII_MGMT_CFG] |= TSI108_MAC_MII_MGMT_RST;
	udelay(100);
	data->regs[TSI108_MAC_MII_MGMT_CFG] &= ~(TSI108_MAC_MII_MGMT_RST |TSI108_MAC_MII_MGMT_CLK);
	data->regs[TSI108_MAC_MII_MGMT_CFG] |= TSI108_MAC_MII_MGMT_CLK;
}



static int tsi108_set_mac(struct eth_device *dev, void *addr)
{
	drv_data *data = dev->priv;
	u32 word1, word2;
	char *str, *eth, buf[64];
	int   devnum = -1;
	int x;

	/* Get MAC from environment */
	devnum = (int)dev->name[(strlen(dev->name) - 1)] - 0x30; /* convert ascii to int */

	switch(devnum)  {
	case 0:
	  str = "ethaddr";
	  break;
	case 1:
	  str = "eth1addr";
	  break;
	default:
	  printf ("$s: Invalid device number %d\n", __FUNCTION__, devnum);
	  return 1;  
	}

	/* Get etheraddr from environment */
	str = (getenv_r (str, buf, sizeof (buf)) > 0) ? buf : NULL;

	if (str) {
		for (x=0; x < 6; ++x) {
			dev->enetaddr[x] = str ? simple_strtoul (str, &eth, 16) : 0;	/* cycle through octets */
			
			if (str)
				str = (*eth) ? eth + 1 : eth;
		}

		word1 = (dev->enetaddr[5] << 24) |
		  (dev->enetaddr[4] << 16) |
		  (dev->enetaddr[3] << 8) |
		  dev->enetaddr[2];

		word2 = (dev->enetaddr[1] << 24) |
		  dev->enetaddr[0] << 16;
		  
	} else {
		/* If not found, use default */
		memcpy((char *)dev->enetaddr, (char *)addr, 6);

			word2 = (*((char *)addr + 1) <<  24)
			  | (*((char *)addr) <<  16);

			word1 = (*((char *)addr + 5) <<  24)
			  | (*((char *)addr + 4) << 16)
			  | (*((char *)addr + 3) << 8)
			  | *((char *)addr + 2);
	}

	data->regs[TSI108_MAC_ADDR1] = word1;
	data->regs[TSI108_MAC_ADDR2] = word2;
	
	data->mac_ok = 1;
	
	return 0;
}



static void tsi108_init_phy(struct eth_device *dev)
{
	drv_data *data = dev->priv;
	u16 phyVal = 0;

	data->regs[TSI108_MAC_IFCTRL] |= TSI108_MAC_IFCTRL_PHYMODE;

	tsi108_write_mii(data, PHY_CTRL, PHY_CTRL_RESET);
	while (tsi108_read_mii(data, PHY_CTRL, NULL) & PHY_CTRL_RESET);

#ifdef CONFIG_TSI108EVB
	tsi108_write_mii(data, 0x09, 0x0300);
	tsi108_write_mii(data, 0x10, 0x1020);
	tsi108_write_mii(data, 0x1C, 0x8C00);
#endif

	tsi108_write_mii(data, PHY_CTRL, PHY_CTRL_AUTONEG_EN | PHY_CTRL_AUTONEG_START);

	sync();
	
	while ((phyVal = tsi108_read_mii(data, PHY_CTRL, NULL)) & PHY_CTRL_AUTONEG_START);

#ifdef ETH_DBG
	printf("phyVal for PHY_CTRL in tsi108_init_phy is %04x\n", phyVal);
#endif
 
	// Set G/MII mode and receive clock select in TBI control #2.  The
	// second port won't work if this isn't done, even though we don't
	// use TBI mode.

	tsi108_write_tbi(data, 0x11, 0x30);

	//FIXME: It seems to take more than 2 back-to-back reads to the
	//  PHY_STAT register before the link up status bit is set.
	u32 i = 0;
	while (!((phyVal = tsi108_read_mii(data, PHY_STAT, NULL)) & PHY_STAT_LINKUP)) {
	  if (i++ > MAX_MII_READS)
	    break;
	}
	
	data->phy_ok = 1; 
	data->link_up = 0;

}

static void tsi108_kill_phy(struct eth_device *dev)
{
  drv_data *data = dev->priv;

  tsi108_write_mii(data, PHY_CTRL, PHY_CTRL_POWERDOWN);
  data->phy_ok = 0;

}

static int tsi108_ifup(struct eth_device *dev)
{
	int i;
	drv_data *data = dev->priv;

#ifdef ETH_DBG
	printf("In tsi108_ifup\n");
#endif
	
	data->rxring = malloc((TSI108_RXRING_LEN + 1) * sizeof(rx_desc));
	if (!data->rxring)
		return -ENOMEM;

	if ((u32)(data->rxring) & 0x1f) {
	  data->rxring = (rx_desc *)(((u32)(data->rxring) + 32) & ~0x1f);
	}

	data->txring = malloc((TSI108_TXRING_LEN + 1) * sizeof(tx_desc));
	if (!data->txring) {
	  free(data->rxring);
	  return -ENOMEM;
	}

	if ((u32)(data->txring) & 0x1f) {
	  data->txring = (tx_desc *)(((u32)(data->txring) + 32) & ~0x1f);
	}

	//FIXME: It seems (from other driver) like malloc is returning
	//  a physical address, if the assumption turns out to be false,
	//  some convertion will have to be done.
	data->rxphys = (u32)data->rxring;
	data->txphys = (u32)data->txring;
#ifdef ETH_DBG
	printf("data->rxphys is %08x\n", data->rxring);
	printf("data->txphys is %08x\n", data->txring);
#endif
	
	for (i = 0; i < TSI108_RXRING_LEN; i++) {
		data->rxring[i].next0 = data->rxphys + (i + 1) * sizeof(rx_desc);
		data->rxring[i].blen = TSI108_RXBUF_SIZE;
		data->rxring[i].vlan = 0;
	}
	
	data->rxring[TSI108_RXRING_LEN - 1].next0 = data->rxphys;

	data->rxtail = 0;
	data->rxhead = 0;

	for (i = 0; i < TSI108_RXRING_LEN; i++) {
	    unsigned char * rxbuff = malloc(TSI108_RXBUF_SIZE);

		if (!rxbuff) {
			printf("Could only allocate %d receive buffer(s).\n", i);
			data->rxhead = i;
			break;
		}

		data->rx_buff[i] = rxbuff;
		data->rxring[i].buf0 = (u32)rxbuff;
		data->rxring[i].misc = TSI108_RX_OWN | TSI108_RX_INT;
	}

	sync();

	data->rxfree = i;
	data->regs[TSI108_EC_RXQ_PTRLOW] = data->rxphys;

	for (i = 0; i < TSI108_TXRING_LEN; i++) {
		// tx buffers do not have to be allocated now
		data->txring[i].next0 = data->txphys + (i + 1) * sizeof(tx_desc);
		data->txring[i].misc = 0;
	}

	data->txring[TSI108_TXRING_LEN - 1].next0 = data->txphys;

	data->txtail = 0;
	data->txhead = 0;
	data->txfree = TSI108_TXRING_LEN;

	data->regs[TSI108_EC_TXQ_PTRLOW] = data->txphys;

	tsi108_init_phy(dev);
	tsi108_check_phy(dev);
	tsi108_restart_rx(data, dev);

	data->regs[TSI108_EC_INTSTAT] = ~0;
	sync();
	data->regs[TSI108_EC_INTMASK] = ~(TSI108_INT_TXERROR | TSI108_INT_TXIDLE |
	                              TSI108_INT_RXERROR | TSI108_INT_RXTHRESH |
	                              TSI108_INT_RXQUEUE0 | TSI108_INT_RXOVERRUN |
	                              TSI108_INT_RXWAIT | TSI108_INT_SFN |
	                              TSI108_INT_STATCARRY);

	data->regs[TSI108_MAC_CFG1] = TSI108_MAC_CFG1_RXEN | TSI108_MAC_CFG1_TXEN;

	return 0;
}

static void tsi108_init_mac(struct eth_device *dev)
{
	drv_data *data = dev->priv;

	data->regs[TSI108_MAC_CFG2] = TSI108_MAC_CFG2_DFLT_PREAMBLE | TSI108_MAC_CFG2_PADCRC;
	
	data->regs[TSI108_EC_TXTHRESH] = (192 << TSI108_EC_TXTHRESH_STARTFILL)
	                           | (192 << TSI108_EC_TXTHRESH_STOPFILL);

	data->regs[TSI108_STAT_CARRYMASK1] = ~(TSI108_STAT_CARRY1_RXBYTES |
	                                   TSI108_STAT_CARRY1_RXPKTS |
	                                   TSI108_STAT_CARRY1_RXFCS |
	                                   TSI108_STAT_CARRY1_RXMCAST |
	                                   TSI108_STAT_CARRY1_RXALIGN |
	                                   TSI108_STAT_CARRY1_RXLENGTH |
	                                   TSI108_STAT_CARRY1_RXRUNT |
	                                   TSI108_STAT_CARRY1_RXJUMBO |
	                                   TSI108_STAT_CARRY1_RXFRAG |
	                                   TSI108_STAT_CARRY1_RXJABBER |
	                                   TSI108_STAT_CARRY1_RXDROP);

	data->regs[TSI108_STAT_CARRYMASK2] = ~(TSI108_STAT_CARRY2_TXBYTES |
	                                   TSI108_STAT_CARRY2_TXPKTS |
	                                   TSI108_STAT_CARRY2_TXEXDEF |
	                                   TSI108_STAT_CARRY2_TXEXCOL |
	                                   TSI108_STAT_CARRY2_TXTCOL |
	                                   TSI108_STAT_CARRY2_TXPAUSE);

	data->regs[TSI108_EC_PORTCTRL] = TSI108_EC_PORTCTRL_STATEN;
	data->regs[TSI108_MAC_CFG1] = 0;
	data->regs[TSI108_MAC_CFG1] = TSI108_MAC_CFG1_RXEN | TSI108_MAC_CFG1_TXEN;
	data->regs[TSI108_EC_RXCFG] = TSI108_EC_RXCFG_SE | TSI108_EC_RXCFG_BFE | TSI108_EC_RXCFG_MFE;

	data->regs[TSI108_EC_TXQ_CFG] = TSI108_EC_TXQ_CFG_DESC_INT |
	  TSI108_EC_TXQ_CFG_EOQ_OWN_INT |
	  TSI108_EC_TXQ_CFG_WSWP |
	  (TSI108_PBM_PORT << TSI108_EC_TXQ_CFG_SFNPORT);

	data->regs[TSI108_EC_RXQ_CFG] = TSI108_EC_RXQ_CFG_DESC_INT |
	  TSI108_EC_RXQ_CFG_EOQ_OWN_INT |
	  TSI108_EC_RXQ_CFG_WSWP |
	  (TSI108_PBM_PORT << TSI108_EC_RXQ_CFG_SFNPORT);

	data->regs[TSI108_EC_TXQ_BUFCFG] = TSI108_EC_TXQ_BUFCFG_BURST256 |
	  TSI108_EC_TXQ_BUFCFG_BSWP |
	  (TSI108_PBM_PORT << TSI108_EC_TXQ_BUFCFG_SFNPORT);
	
	data->regs[TSI108_EC_RXQ_BUFCFG] = TSI108_EC_RXQ_BUFCFG_BURST256 |
	  TSI108_EC_RXQ_BUFCFG_BSWP |
	  (TSI108_PBM_PORT << TSI108_EC_RXQ_BUFCFG_SFNPORT);

	data->regs[TSI108_EC_INTMASK] = ~0;
}

static int tsi108_init_one(int port, unsigned long regs, unsigned long phyregs, int phy_addr)
{
  struct eth_device *dev = tsi108_devs[port];
  drv_data *data = dev->priv = malloc(sizeof(drv_data));
  
	if (!data) {
		printf("tsi108_eth: Could not allocate driver data structure.\n");
		return -ENOMEM;
	}
	
	data->regs = (volatile u32 *)regs;
	data->phyregs = (volatile u32 *)phyregs;
	data->phy = phy_addr;

	tsi108_reset_ether(data);
	tsi108_kill_phy(dev);
	
	tsi108_set_mac(dev, &(mac_addr[port][0]));
	
	tsi108_init_mac(dev);
	tsi108_ifup(dev);

	tsi108_devs[port] = dev;

	return 0;
}



static int tsi108_ether_init(struct eth_device *dev, bd_t *bis)
{
	int ret = 0;
	static unsigned char memAllocated = 0;

	if (memAllocated) return 1;
#ifdef ETH_DBG
	printf("In tsi108_ether_init!\n");
#endif
	
	if (strcmp(dev->name, "TSI108_GE_PORT0") == 0) {
	  ret = tsi108_init_one(0, TSI108_GIGE_BASE, TSI108_GIGE_BASE, TSI108_PHY0_ADDR);
	}
	else {
	  printf("Initializing second GigE port!\n");
	  ret = tsi108_init_one(1, TSI108_GIGE_BASE + 0x400, TSI108_GIGE_BASE, TSI108_PHY1_ADDR);
	}
	memAllocated = 1;
	
	if (ret > 0) return 0;
	
	return 1; //success
}

int tsi108_eth_initialize(bd_t *bis)
{
	struct eth_device *dev[TSI108_MAX_NUM_ETH_PORTS];
	int i;

	printf("Initializing TSI108 ethernet ports!\n");
	
	for (i = 0; i < TSI108_MAX_NUM_ETH_PORTS; i++)
	{
	  if (tsi108_devs[i] == NULL) {
	    dev[i] = (struct eth_device *)malloc(sizeof(struct eth_device));
	  }
	  else {
	    dev[i] = tsi108_devs[i];
	  }
		
		dev[i]->init = tsi108_ether_init;
		//FIXME: revisit halt later, may have to disable phy too
		dev[i]->halt = tsi108_stop_ethernet;
		dev[i]->send = tsi108_send_packet;
		dev[i]->recv = tsi108_poll;

		sprintf(dev[i]->name, "TSI108_GE_PORT%d", i);

		tsi108_devs[i] = dev[i];

		eth_register(dev[i]);
	}
	printf("Exiting tsi108_eth_initialize! i = %d\n", i);
	return i;
}
























